
import numpy as np
cimport numpy as np


def cython_fit_BlackScholesMerton_model(np.ndarray ts, np.ndarray prices):
    cdef np.ndarray dlogS = np.log(prices[1:] / prices[:-1])
    cdef np.ndarray dt = ts[1:] - ts[:-1]

    cdef double r = np.mean(dlogS / dt)
    cdef double sigma = np.std(dlogS / np.sqrt(dt))

    return r, sigma


def cython_fit_multivariate_BlackScholesMerton_model(np.ndarray ts, np.ndarray multiprices):
    cdef np.ndarray dlogS = np.log(multiprices[:, 1:] / multiprices[:, :-1])
    cdef np.ndarray dt = ts[1:] - ts[:-1]

    cdef np.ndarray r = np.zeros(multiprices.shape[0])
    for i in range(multiprices.shape[0]):
        r[i] = np.mean((dlogS / dt)[i, :])
    cdef np.ndarray cov = np.cov(dlogS / np.sqrt(dt))

    return r, cov




import unittest
import os

import numpy as np

from finsim.portfolio import get_optimized_portfolio_on_sharpe_ratio, get_optimized_portfolio_on_mpt_costfunction
from finsim.portfolio import OptimizedPortfolio, Portfolio


class TestPortfolio(unittest.TestCase):
    def test_modern_portfolio_sharpe_ratio(self):
        selected_symbols = ['NVDA', 'AMZN', 'BAH', 'GD']

        startdate = '2010-01-01'
        enddate = '2020-03-31'

        optimized_portfolio = get_optimized_portfolio_on_sharpe_ratio(0.0123, selected_symbols, 1000,
                                                                      enddate, startdate, enddate,
                                                                      minweight=0.2)

        summary = optimized_portfolio.summary
        
        self.assertAlmostEqual(summary['yield'], 0.2837451882660389, places=4)
        self.assertAlmostEqual(summary['volatility'], 0.24625264201343075, places=4)
        self.assertAlmostEqual(summary['sharpe_ratio'], 1.1023037433624268, places=4)
        np.testing.assert_array_almost_equal(
            summary['correlation'],
            np.array([[1., 0.38132613, 0.28049998, 0.42612949],
                      [0.38132613, 1., 0.19692159, 0.37066551],
                      [0.28049998, 0.19692159, 1., 0.34781397],
                      [0.42612949, 0.37066551, 0.34781397, 1.]]
                     ),
            decimal=4
        )

        simplified_portfolio = optimized_portfolio.get_portfolio()
        self.assertTrue(isinstance(simplified_portfolio, Portfolio))
        self.assertFalse(isinstance(simplified_portfolio, OptimizedPortfolio))

    def test_modern_postfolio_costfunction(self):
        selected_symbols = ['NVDA', 'AMZN', 'BAH', 'GD']

        startdate = '2010-01-01'
        enddate = '2020-03-31'

        optimized_portfolio = get_optimized_portfolio_on_mpt_costfunction(0.0123, selected_symbols, 1000,
                                                                          enddate, startdate, enddate,
                                                                          0.1)

        summary = optimized_portfolio.summary

        self.assertAlmostEqual(summary['yield'], 0.2837451882660389, places=4)
        self.assertAlmostEqual(summary['volatility'], 0.24625264201343075, places=4)
        self.assertAlmostEqual(summary['mpt_costfunction'], 2.1005125045776367, places=4)
        np.testing.assert_array_almost_equal(
            summary['correlation'],
            np.array([[1.        , 0.38132613, 0.28049998, 0.42612949],
                       [0.38132613, 1.        , 0.19692159, 0.37066551],
                       [0.28049998, 0.19692159, 1.        , 0.34781397],
                       [0.42612949, 0.37066551, 0.34781397, 1.        ]]
                     ),
            decimal=4
        )
        simplified_portfolio = optimized_portfolio.get_portfolio()
        self.assertTrue(isinstance(simplified_portfolio, Portfolio))
        self.assertFalse(isinstance(simplified_portfolio, OptimizedPortfolio))


        optimized_portfolio.save_to_json(open('portfolio.json', 'w'))
        reloaded_portfolio = Portfolio.load_from_json(open('portfolio.json', 'r'))
        self.assertAlmostEqual(
            optimized_portfolio.get_portfolio_value('2020-03-31'),
            reloaded_portfolio.get_portfolio_value('2020-03-31')
        )
        self.assertTrue(isinstance(reloaded_portfolio, Portfolio))
        self.assertFalse(isinstance(reloaded_portfolio, OptimizedPortfolio))
        os.remove('portfolio.json')


if __name__ == '__main__':
    unittest.main()




import unittest
import logging

from finsim.portfolio import DynamicPortfolio, InsufficientSharesException

logging.basicConfig(level=logging.DEBUG)

class TestDynamicPortfolio(unittest.TestCase):
    def setUp(self):
        self.dynport = DynamicPortfolio({'NVDA': 1}, '2020-01-01')
        self.dynport.trade('2020-02-01', buy_stocks={'NVDA': 2, 'VOO': 1})
        self.dynport.trade('2020-03-01', buy_stocks={'LMT': 1}, sell_stocks={'NVDA': 3})
        self.dynport.trade('2020-04-01', buy_stocks={'COF': 2}, sell_stocks={'LMT': 2})

    def test_orders(self):
        self.assertTrue(self.dynport.is_sorted())
        self.assertEqual(self.dynport.find_cursor_for_date('2018-12-15'), -1)
        self.assertEqual(self.dynport.find_cursor_for_date('2020-01-06'), 0)
        self.assertEqual(self.dynport.find_cursor_for_date('2020-02-14'), 1)
        self.assertEqual(self.dynport.find_cursor_for_date('2020-03-22'), 2)
        self.assertEqual(self.dynport.find_cursor_for_date('2020-04-28'), 3)

    def test_timevalue(self):
        self.assertTrue(self.dynport.get_portfolio_value('2020-03-02'),
                        self.dynport.timeseries[2]['portfolio'].get_portfolio_value('2020-03-02'))

    def test_negative_shares(self):
        self.assertLess(self.dynport.timeseries[3]['portfolio'].symbols_nbshares['LMT'], 0)

    def test_trading_notcheck(self):
        port1 = DynamicPortfolio({'NVDA': 2, 'MSFT': 3}, '2020-01-01')
        port1.trade('2020-01-15', buy_stocks={'LMT': 3, 'NVDA': 1}, sell_stocks={'MSFT': 10})
        self.assertEqual(port1.symbols_nbshares['NVDA'], 3)
        self.assertEqual(port1.symbols_nbshares['LMT'], 3)
        self.assertEqual(port1.symbols_nbshares['MSFT'], -7)

    def test_trading_check_available_fund(self):
        port2 = DynamicPortfolio({'NVDA': 2, 'MSFT': 3}, '2020-01-01')
        port2.trade('2020-01-15', buy_stocks={'LMT': 3, 'NVDA': 1}, sell_stocks={'MSFT': 10},
                    check_valid=True)
        self.assertEqual(port2.symbols_nbshares['NVDA'], 3)
        self.assertEqual(port2.symbols_nbshares['LMT'], 3)
        self.assertEqual(port2.symbols_nbshares['MSFT'], 0)

    def test_trading_raise_exception(self):
        port3 = DynamicPortfolio({'NVDA': 2, 'MSFT': 3}, '2020-01-01')
        with self.assertRaises(InsufficientSharesException):
            port3.trade('2020-01-15', buy_stocks={'LMT': 3, 'NVDA': 1}, sell_stocks={'MSFT': 10},
                        check_valid=True, raise_insufficient_stock_error=True)


if __name__ == '__main__':
    unittest.main()




import json
import logging
from collections import defaultdict

from tqdm import tqdm
import pandas as pd

from ..data.preader import get_yahoofinance_data, get_symbol_closing_price


class Portfolio:
    def __init__(self, symbols_nbshares, cacheddir=None):   # symbols_nbshares = {'NVDA': 200, 'AMZN': 101}
        self.symbols_nbshares = symbols_nbshares
        self.cacheddir = cacheddir

    def get_portfolio_value(self, datestr):
        portfolio_value = sum([
            self.symbols_nbshares[symbol] * get_symbol_closing_price(symbol, datestr, cacheddir=self.cacheddir)
            for symbol in self.symbols_nbshares
        ])
        return portfolio_value

    def get_portfolio_values_overtime(self, startdate, enddate, cacheddir=None):
        logging.info('Reading financial data...')
        stocks_data_dfs = [
            get_yahoofinance_data(sym, startdate, enddate, cacheddir=cacheddir)
            for sym in tqdm(self.symbols_nbshares.keys())
        ]

        logging.info('Estimating...')
        max_timearray_ref = 0
        maxlen = max(len(stocks_data_dfs[i]) for i in range(len(stocks_data_dfs)))
        minlen = min(len(stocks_data_dfs[i]) for i in range(len(stocks_data_dfs)))
        if minlen != maxlen:
            logging.warning('Not all symbols have data all the way back to {}'.format(startdate))
            max_timearray_ref = [i for i in range(len(stocks_data_dfs)) if maxlen == len(stocks_data_dfs[i])][0]
            logging.warning('Symbols not having whole range of data:')
            for i, symbol in enumerate(self.symbols_nbshares):
                thisdflen = len(stocks_data_dfs[i])
                if thisdflen != maxlen:
                    if thisdflen == 0:
                        logging.warning('No data for {} for this date range at all.'.format(symbol))
                        predf = pd.DataFrame(stocks_data_dfs[max_timearray_ref]['TimeStamp'])
                    else:
                        logging.warning('{}: starting from {}'.format(symbol, stocks_data_dfs[i]['TimeStamp'][0].date().strftime('%Y-%m-%d')))
                        predf = pd.DataFrame(stocks_data_dfs[max_timearray_ref]['TimeStamp'][:(maxlen - thisdflen)])
                    predf[stocks_data_dfs[max_timearray_ref].columns[1:]] = 0
                    stocks_data_dfs[i] = predf.append(stocks_data_dfs[i])

        df = pd.DataFrame(stocks_data_dfs[max_timearray_ref]['TimeStamp'].copy())
        df['value'] = sum([
            self.symbols_nbshares[sym] * stocks_data_dfs[i]['Close']
            for i, sym in enumerate(self.symbols_nbshares.keys())
        ])
        return df

    def __add__(self, other):
        assert isinstance(other, Portfolio)

        symshares1 = defaultdict(lambda: 0, self.symbols_nbshares)
        symshares2 = defaultdict(lambda: 0, other.symbols_nbshares)
        all_symbols = set(symshares1.keys()).union(symshares2.keys())

        total_symbols_shares = {symbol: symshares2[symbol] + symshares1[symbol] for symbol in all_symbols}
        return Portfolio(total_symbols_shares, cacheddir=self.cacheddir)

    def __sub__(self, other):
        assert isinstance(other, Portfolio)

        symshares1 = defaultdict(lambda: 0, self.symbols_nbshares)
        symshares2 = defaultdict(lambda: 0, other.symbols_nbshares)
        all_symbols = set(symshares1.keys()).union(symshares2.keys())

        symbols_diff_shares = {symbol: symshares1[symbol] - symshares2[symbol] for symbol in all_symbols}
        return Portfolio(symbols_diff_shares, cacheddir=self.cacheddir)

    def __eq__(self, other):
        assert isinstance(other, Portfolio)

        for symbol, nbshares in self.symbols_nbshares.items():
            if symbol not in other.symbols_nbshares:
                return False
            if other.symbols_nbshares[symbol] != nbshares:
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def portfolio_symbols_nbshares(self):
        return self.symbols_nbshares

    def roundoff_nbshares(self):
        for symbol in self.symbols_nbshares:
            nbshares = round(self.symbols_nbshares[symbol])
            self.symbols_nbshares[symbol] = nbshares

    def multiply(self, factor):
        for symbol in self.symbols_nbshares:
            nbshares = self.symbols_nbshares[symbol]
            self.symbols_nbshares[symbol] = nbshares * factor
            
    def __mul__(self, other):
        assert isinstance(other, int) or isinstance(other, float)
        newshares = {symbol: nbshares*other for symbol, nbshares in self.symbols_nbshares.items()}
        return Portfolio(newshares, cacheddir=self.cacheddir)

    def save_to_json(self, fileobj):
        json.dump(self.symbols_nbshares, fileobj)

    def dumps_json(self):
        return json.dumps(self.symbols_nbshares)

    @classmethod
    def load_from_json(cls, fileobj, cacheddir=None):
        symbols_nbshares = json.load(fileobj)
        return cls(symbols_nbshares, cacheddir=cacheddir)

    @classmethod
    def load_from_dict(cls, portdict, cacheddir=None):
        return cls(portdict, cacheddir=cacheddir)


class OptimizedPortfolio(Portfolio):
    def __init__(self, policy, totalworth, presetdate, cacheddir=None):
        super(OptimizedPortfolio, self).__init__({}, cacheddir=cacheddir)
        self.policy = policy
        self.totalworth = totalworth
        self.presetdate = presetdate
        self.compute()

    def compute(self):
        prices = {
            symbol: get_symbol_closing_price(symbol, self.presetdate, cacheddir=self.cacheddir)
            for symbol in self.policy.symbols
        }
        summary = self.policy.portfolio_summary
        for component in summary['components']:
            symbol = component['symbol']
            component['nbshares'] = component['weight'] * self.totalworth / prices[symbol]
            self.symbols_nbshares[symbol] = component['nbshares']

        self.summary = summary

    @property
    def portfolio_symbols(self):
        return self.policy.portfolio_symbols

    @property
    def weights(self):
        return self.policy.weights

    @property
    def portfolio_yield(self):
        return self.policy.portfolio_yield

    @property
    def volatility(self):
        return self.policy.volatility

    @property
    def correlation_matrix(self):
        return self.policy.correlation_matrix

    @property
    def named_correlation_matrix(self):
        return self.policy.named_correlation_matrix

    @property
    def portfolio_summary(self):
        return self.summary

    def get_portfolio(self):
        return Portfolio(self.symbols_nbshares, cacheddir=self.cacheddir)

    def save_to_json(self, fileobj):
        self.get_portfolio().save_to_json(fileobj)

    def dumps_json(self):
        return self.get_portfolio().dumps_json()

    @classmethod
    def load_from_json(cls, fileobj, cacheddir=None):
        raise NotImplementedError('OptimizedPortfolio does not implement loading from json. Use Portfolio to load instead.')

    @classmethod
    def load_from_dict(cls, portdict, cacheddir=None):
        raise NotImplementedError('OptimizedPortfolio does not implement loading from json. Use Portfolio to load instead.')




import numpy as np
cimport numpy as np


def cython_sharpe_ratio(np.ndarray weights, np.ndarray r, np.ndarray cov, double rf):
    cdef double yieldrate = np.sum(weights * r)
    cdef np.ndarray sqweights = np.expand_dims(weights, axis=1) @ np.expand_dims(weights, axis=0)
    cdef double volatility = np.sqrt(np.sum(sqweights * cov))
    return (yieldrate - rf) / volatility


def cython_mpt_costfunction(np.ndarray weights, np.ndarray r, np.ndarray cov, double rf, double lamb, double V0):
    cdef np.ndarray weightmat = np.expand_dims(weights[:-1], axis=0)
    cdef double c = lamb * V0
    cdef double cost = weights[-1]*rf + np.dot(weights[:-1], r) - 0.5*c/V0*(weightmat @ cov @ weightmat.T)[0, 0]
    return cost


def cython_mpt_entropy_costfunction(np.ndarray weights, np.ndarray r, np.ndarray cov, double rf, double lamb0, double lamb1, double V):
    cdef np.ndarray weightmat = np.expand_dims(weights[:-1], axis=0)
    cdef double c0 = lamb0 * V
    cdef double c1 = lamb1 * V
    cdef double yield_val = weights[-1]*rf + np.dot(weights[:-1], r)
    cdef double cov_val = - 0.5 * c0 / V * (weightmat @ cov @ weightmat.T)[0, 0]
    cdef double sumweights = np.sum(weights[:-1])
    cdef double entropy_val = - 0.5 * c1 / V * np.sum(weights[:-1] * (np.log(weights[:-1]) - np.log(sumweights))) / sumweights
    return yield_val + cov_val + entropy_val




from datetime import datetime, timedelta
from operator import itemgetter
from collections import defaultdict
import logging
import json

import numpy as np
import pandas as pd
from .portfolio import Portfolio
from ..data.preader import get_dividends_df


class InsufficientSharesException(Exception):
    pass


class DynamicPortfolio(Portfolio):
    def __init__(self, symbol_nbshares, current_date, cacheddir=None):
        # current_date is a string, of format '%Y-%m-%d', such as '2020-02-23'
        super(DynamicPortfolio, self).__init__(symbol_nbshares, cacheddir=cacheddir)
        self.current_date = current_date
        self.timeseries = [{
            'date': self.current_date,
            'portfolio': Portfolio(self.symbols_nbshares, self.cacheddir)
        }]
        self.timeseriesidx = 0

    def sort_time_series(self):
        self.timeseries = sorted(self.timeseries, key=itemgetter('date'))

    def is_sorted(self):
        for i, j in zip(range(len(self.timeseries)-1), range(1, len(self.timeseries))):
            if not self.timeseries[i]['date'] < self.timeseries[j]['date']:
                return False
        return True

    def find_cursor_for_date(self, date):
        # date is a string, of format '%Y-%m-%d', such as '2020-02-23'
        start = 0
        end = len(self.timeseries)
        idx = (end-start) // 2

        if date < self.timeseries[0]['date']:
            return (-1)
        if date >= self.timeseries[end-1]['date']:
            return (end-1)

        found = False
        while not found:
            if date >= self.timeseries[idx]['date'] and (idx >= len(self.timeseries) or date < self.timeseries[idx+1]['date']):
                found = True
            else:
                if self.timeseries[idx]['date'] < date:
                    start = idx
                elif self.timeseries[idx]['date'] >= date:
                    end = idx
                idx = (start+end) // 2
        return idx

    def move_cursor_to_date(self, newdate):
        # date is a string, of format '%Y-%m-%d', such as '2020-02-23'
        self.timeseriesidx = self.find_cursor_for_date(newdate)
        self.symbols_nbshares = self.timeseries[self.timeseriesidx]['portfolio'].symbols_nbshares.copy()
        self.current_date = newdate

    def trade(
            self,
            trade_date,
            buy_stocks={},
            sell_stocks={},
            check_valid=False,
            raise_insufficient_stock_error=False
    ):
        assert self.is_sorted()
        assert trade_date > self.timeseries[-1]['date']

        # rectifying
        overlapping_symbols = set(buy_stocks.keys()).intersection(set(sell_stocks.keys()))
        for symbol in overlapping_symbols:
            if buy_stocks[symbol] > sell_stocks[symbol]:
                buy_stocks[symbol] -= sell_stocks[symbol]
                del sell_stocks[symbol]
            elif buy_stocks[symbol] < sell_stocks[symbol]:
                sell_stocks[symbol] -= buy_stocks[symbol]
                del buy_stocks[symbol]
            else:
                del buy_stocks[symbol]
                del sell_stocks[symbol]

        symbols_nbshares = defaultdict(lambda : 0, buy_stocks)
        if check_valid:
            if raise_insufficient_stock_error:
                errorfound = False
                for symbol, nbshares in sell_stocks.items():
                    if nbshares > self.symbols_nbshares[symbol]:
                        logging.error('Insufficient stock {} ({} shares, but want to sell {} shares)'.format(symbol, self.symbols_nbshares[symbol], nbshares))
                        errorfound = True
                if errorfound:
                    raise InsufficientSharesException('Insufficient stocks.')
            else:
                for symbol, nbshares in sell_stocks.items():
                    if nbshares > self.symbols_nbshares[symbol]:
                        sell_stocks[symbol] = self.symbols_nbshares[symbol]

        for symbol, nbshares in sell_stocks.items():
            symbols_nbshares[symbol] -= nbshares

        difference_portfolio = Portfolio(symbols_nbshares, cacheddir=self.cacheddir)
        self.timeseries.append({
            'date': trade_date,
            'portfolio': self.timeseries[-1]['portfolio'] + difference_portfolio
        })

        self.move_cursor_to_date(trade_date)
        
    def get_portfolio_value(self, datestr):
        idx = self.find_cursor_for_date(datestr)
        return self.timeseries[idx]['portfolio'].get_portfolio_value(datestr)

    def get_portfolio_values_overtime(self, startdate, enddate, cacheddir=None):
        assert self.is_sorted()

        startidx = self.find_cursor_for_date(startdate)
        endidx = self.find_cursor_for_date(enddate)

        assert startidx <= endidx

        dataframes = []
        for portidx in range(startidx, endidx+1):
            thisport_startdate = max(startdate, self.timeseries[portidx]['date'])
            if portidx == endidx:
                thisport_enddate = enddate
            else:
                ed = datetime.strptime(self.timeseries[portidx+1]['date'], '%Y-%m-%d') - timedelta(1)
                thisport_enddate = datetime.strftime(ed, '%Y-%m-%d')
            dataframes.append(
                self.timeseries[portidx]['portfolio'].get_portfolio_values_overtime(
                    thisport_startdate,
                    thisport_enddate,
                    cacheddir=self.cacheddir
                )
            )

        df = pd.concat(dataframes)
        df['TimeStamp'] = df['TimeStamp'].map(lambda ts: datetime.strftime(ts, '%Y-%m-%d'))
        return df

    def generate_dynamic_portfolio_dict(self):
        dynport_dict = {'name': 'DynamicPortfolio'}
        dynport_dict['current_date'] = self.current_date
        dynport_dict['timeseries'] = []
        for portdict in self.timeseries:
            trade_date = portdict['date']
            portfolio = portdict['portfolio']
            dynport_dict['timeseries'].append({
                'date': trade_date,
                'portfolio': portfolio.symbols_nbshares
            })
        return dynport_dict

    def save_to_json(self, fileobj):
        dynport_dict = self.generate_dynamic_portfolio_dict()
        json.dump(dynport_dict, fileobj)

    def dumps_json(self):
        dynport_dict = self.generate_dynamic_portfolio_dict()
        return json.dumps(dynport_dict)

    @classmethod
    def load_from_dict(cls, dynportdict, cacheddir=None):
        assert dynportdict['name'] == 'DynamicPortfolio'
        dynport = cls(
            dynportdict['timeseries'][0]['portfolio'],
            dynportdict['timeseries'][0]['date'],
            cacheddir=cacheddir
        )
        for portdict in dynportdict['timeseries'][1:]:
            tradedate = portdict['date']
            symbols_nbshares = portdict['portfolio']
            dynport.timeseries.append({'date': tradedate,
                                       'portfolio': Portfolio(symbols_nbshares, cacheddir=cacheddir)
                                       })

        dynport.move_cursor_to_date(dynportdict['current_date'])
        return dynport
    
    @classmethod
    def load_from_json(cls, fileobj, cacheddir=None):
        dynportinfo = json.load(fileobj)
        return cls.load_from_dict(dynportinfo, cacheddir=cacheddir)


class DynamicPortfolioWithDividends(DynamicPortfolio):
    def __init__(self, symbol_nbshares, current_date, cash=0., cacheddir=None):
        # current_date is a string, of format '%Y-%m-%d', such as '2020-02-23'
        super(DynamicPortfolioWithDividends, self).__init__(symbol_nbshares, current_date, cacheddir=cacheddir)
        self.startcash = cash
        self.cashtimeseries = [{'date': current_date, 'cash': self.startcash}]

    def calculate_cash_from_dividends(self, enddate):
        startdate = self.timeseries[0]['date']
        self.cashtimeseries = [{'date': startdate, 'dividend': 0., 'cash': self.startcash}]

        for i, periodinfo in enumerate(self.timeseries):
            series_startdate = periodinfo['date']
            series_enddate = self.timeseries[i+1]['date'] if i < len(self.timeseries)-1 else enddate

            symbol_dividends_dataframes = []
            for symbol, nbshares in periodinfo['portfolio'].symbols_nbshares.items():
                dividend_df = get_dividends_df(symbol)
                if len(dividend_df) == 0:
                    continue
                dividend_df = dividend_df[(dividend_df['TimeStamp']>=series_startdate) & (dividend_df['TimeStamp']<series_enddate)].copy()
                if len(dividend_df) == 0:
                    continue
                dividend_df['Dividends'] *= nbshares
                symbol_dividends_dataframes.append(dividend_df)

            if len(symbol_dividends_dataframes) == 0:
                continue

            period_dividend_dataframe = pd.concat(symbol_dividends_dataframes)
            period_dividend_dataframe = period_dividend_dataframe.sort_values('TimeStamp')

            for dividend_date, dividend in zip(period_dividend_dataframe['TimeStamp'], period_dividend_dataframe['Dividends']):
                cash = self.cashtimeseries[-1]['cash']
                self.cashtimeseries.append({'date': dividend_date, 'dividend': dividend, 'cash': cash+dividend})

    def get_portfolio_values_overtime(self, startdate, enddate, cacheddir=None):
        worthdf = super(DynamicPortfolioWithDividends, self).get_portfolio_values_overtime(startdate, enddate, cacheddir=cacheddir)
        worthdf.rename(columns={'value': 'stock_value'}, inplace=True)
        self.calculate_cash_from_dividends(enddate)
        
        dividends_df = pd.DataFrame.from_records(self.cashtimeseries) if len(self.cashtimeseries) > 0 else pd.DataFrame(
            np.empty(0, dtype=[('date', 'S20'), ('dividend', 'f8'), ('cash', 'f8')])
        )
        dividends_df = dividends_df.rename(columns={'date': 'TimeStamp'})
        dividends_df = dividends_df.groupby('TimeStamp').aggregate({'dividend': 'sum', 'cash': 'max'}).reset_index()
        dividends_df = dividends_df.sort_values('TimeStamp')
        dividends_df['TimeStamp'] = dividends_df['TimeStamp'].astype(str)
        worthdf = worthdf.merge(dividends_df, on='TimeStamp', how='left')
        worthdf['cash'] = worthdf['cash'].ffill()
        worthdf = worthdf.fillna(0.)
        worthdf['value'] = worthdf['stock_value'] + worthdf['cash']

        return worthdf




import numpy as np
cimport numpy as np


def cython_estimate_downside_risk(np.ndarray ts, np.ndarray prices, double target_return):
    cdef np.ndarray dlogS = np.log(prices[1:] / prices[:-1])
    cdef np.ndarray dt = ts[1:] - ts[:-1]

    cdef np.ndarray rms_rarray = dlogS / np.sqrt(dt)
    cdef np.ndarray less_return_array = target_return - rms_rarray
    less_return_array[less_return_array < 0] = 0.
    cdef double downside_risk = np.sqrt(np.mean(np.square(less_return_array)))
    return downside_risk


def cython_estimate_upside_risk(np.ndarray ts, np.ndarray prices, double target_return):
    cdef np.ndarray dlogS = np.log(prices[1:] / prices[:-1])
    cdef np.ndarray dt = ts[1:] - ts[:-1]

    cdef np.ndarray rms_rarray = dlogS / np.sqrt(dt)
    cdef np.ndarray more_return_array = rms_rarray - target_return
    more_return_array[more_return_array < 0] = 0.
    cdef double upside_risk = np.sqrt(np.mean(np.square(more_return_array)))
    return upside_risk




import numpy as np


def python_fit_BlackScholesMerton_model(ts, prices):
    dlogS = np.log(prices[1:] / prices[:-1])
    dt = ts[1:] - ts[:-1]

    r = np.mean(dlogS / dt)
    sigma = np.std(dlogS / np.sqrt(dt))

    return r, sigma


def python_fit_multivariate_BlackScholesMerton_model(ts, multiprices):
    dlogS = np.log(multiprices[:, 1:] / multiprices[:, :-1])
    dt = ts[1:] - ts[:-1]

    r = np.zeros(multiprices.shape[0])
    for i in range(multiprices.shape[0]):
        r[i] = np.mean((dlogS / dt)[i, :])
    # r = np.mean((dlogS / dt), axis=1)
    cov = np.cov(dlogS / np.sqrt(dt))

    return r, cov
